"""
Wrappers for analogous functions in scipy.sparse.csgraph
"""

import numpy as np

from scipy.sparse import csr_matrix

from ..dag import geodesic_matrix
from ..dag import connected_components as _connected_components


def dijkstra(
    csgraph,
    directed=True,
    indices=None,
    return_predecessors=False,
    unweighted=False,
    limit=np.inf,
    min_only=False,
    checks=True,
):
    """Wrapper for geodesic_matrix() that mimics the interfaces of scipy.sparse.csgraph.dijkstra().

    Notes:
    1. `min_only=True` is not supported.
    2. `return_predecessors=True` is currently not supported (might change in the future)
    3. `checks` (enabled by default) will raise an error if the input graph is not a tree.

    """
    if min_only or return_predecessors:
        raise NotImplementedError(
            "min_only=True and return_predecessors=True are not supported."
        )

    # If input is not csr matrix, try to convert it
    if not isinstance(csgraph, csr_matrix):
        csgraph = csr_matrix(csgraph)

    # Turn the csr matrix into two numpy arrays: sources and targets
    sources, targets = csgraph.nonzero()

    # If checks, ensure that the input graph is actually a tree
    # This means each node must only have one parent (out-degree <= 1)
    if checks:
        _, counts = np.unique(sources, return_counts=True)
        if np.any(counts > 1):
            raise ValueError(
                "Input graph is not a tree. Each node must have at most one parent."
            )

    dmat = geodesic_matrix(
        sources,
        targets,
        directed=directed,
        indices=indices,
        weights=csgraph.data if not unweighted else None,
    )

    # Djikstra would return unreachable nodes as np.inf
    dmat[dmat < 0] = np.inf

    if limit is not np.inf:
        dmat[dmat > limit] = np.inf

    return dmat


def connected_components(csgraph, directed=True, connection='weak', return_labels=True, checks=True):
    """Wrapper for connected_components() that mimics the interfaces of scipy.sparse.csgraph.connected_components().

    Notes:
    1. `connection='strong'` makes no sense for trees, so it is not supported.

    """
    if connection == 'strong':
        raise NotImplementedError("connection='strong' makes no sense for trees, so it is not supported.")

    # If input is not csr matrix, try to convert it
    if not isinstance(csgraph, csr_matrix):
        csgraph = csr_matrix(csgraph)

    # Turn the csr matrix into two numpy arrays: sources and targets
    sources, targets = csgraph.nonzero()

    # If checks, ensure that the input graph is actually a tree
    # This means each node must only have one parent (out-degree <= 1)
    if checks:
        _, counts = np.unique(sources, return_counts=True)
        if np.any(counts > 1):
            raise ValueError(
                "Input graph is not a tree. Each node must have at most one parent."
            )

    # Get the connected components
    cc = _connected_components(sources, targets)

    # csgraph's connected components returns contiguous labels starting from 0
    # whereas our connectec components returns the ID of the root node
    unique, cc = np.unique(cc, return_inverse=True)

    if not return_labels:
        return len(unique)

    return len(unique), cc


import numpy as np

from collections import namedtuple

from . import _fastcore

__all__ = ["nblast_allbyall"]

Dotprop = namedtuple("Dotprop", ["points", "vect"])


def nblast_allbyall(x, backend="bosque"):
    """All-by-all NBLAST.

    Parameters
    ----------
    x :     list-like with dotprop-likes
            Must have attributes `points` and `vect` that are numpy arrays.

    """
    # TODO:
    # - add support for NBLAST parameters (query, target, scores, precision, etc.)
    # - add support for progress bar
    if not hasattr(x, "__iter__"):
        raise TypeError("x must be iterable")

    for n in x:
        if not hasattr(n, "points") or not hasattr(n, "vect"):
            raise TypeError("x must be iterable of dotprop-likes")

    # Collect points and vectors
    points = [n.points.astype(np.float64, copy=False) for n in x]
    vects = [n.vect.astype(np.float64, copy=False) for n in x]

    # Calculate all-by-all NBLAST
    scores = _fastcore.nblast_allbyall(points, vects, backend=backend)

    return scores


def _make_dotprop(points, k=5):
    """Create a Dotprop object.

    Parameters
    ----------
    points :    np.ndarray
                Array of points.
    k :         int
                Number of nearest neighbours to use for tangent vector
                calculation.

    Returns
    -------
    Dotprop
                Namedtuple with attributes `points` and `vect`.

    """
    vect = _calculate_tangent_vectors(points, k)
    return Dotprop(points, vect)


def _calculate_tangent_vectors(points, k):
    """Calculate tangent vectors.

    Parameters
    ----------
    k :         int
                Number of nearest neighbours to use for tangent vector
                calculation.

    Returns
    -------
    Dotprops
                Only if ``inplace=False``.

    """
    # Create the KDTree and get the k-nearest neighbors for each point
    from scipy.spatial import cKDTree as KDTree

    dist, ix = KDTree(points).query(points, k=k)

    # Get points: array of (N, k, 3)
    pt = points[ix]

    # Generate centers for each cloud of k nearest neighbors
    centers = np.mean(pt, axis=1)

    # Generate vector from center
    cpt = pt - centers.reshape((pt.shape[0], 1, 3))

    # Get inertia (N, 3, 3)
    inertia = cpt.transpose((0, 2, 1)) @ cpt

    # Extract vector and alpha
    u, s, vh = np.linalg.svd(inertia)
    vect = vh[:, 0, :]
    # alpha = (s[:, 0] - s[:, 1]) / np.sum(s, axis=1)

    return vect



import numpy as np

from . import _fastcore

__all__ = [
    "geodesic_matrix",
    "geodesic_pairs",
    "connected_components",
    "synapse_flow_centrality",
    "generate_segments",
    "break_segments",
    "segment_coords",
    "prune_twigs",
    "strahler_index",
    "classify_nodes",
]


def generate_segments(node_ids, parent_ids, weights=None):
    """Generate linear segments maximizing segment lengths.

    Parameters
    ----------
    node_ids :   (N, ) array
                 Array node IDs.
    parent_ids : (N, ) array
                 Array of parent IDs for each node. Root nodes' parents
                 must be -1.
    weights :    (N, ) float32 array, optional
                 Array of distances for each child -> parent connection.
                 If ``None`` all node-to-node distances are set to 1.

    Returns
    -------
    segments :   list of arrays
                 Segments as list of arrays, sorted from longest to shortest.
                 Each segment starts with a leaf and stops with a branch point
                 or root node.

    Examples
    --------
    >>> import navis_fastcore as fastcore
    >>> import numpy as np
    >>> node_ids = np.arange(7)
    >>> parent_ids = np.array([-1, 0, 1, 2, 1, 4, 5])
    >>> fastcore.generate_segments(node_ids, parent_ids)
    [array([6, 5, 4, 1, 0]), array([3, 2, 1])]

    """
    # Convert parent IDs into indices
    parent_ix = _ids_to_indices(node_ids, parent_ids)

    # Make sure weights are float32
    if weights is not None:
        weights = np.asarray(weights, dtype=np.float32, order="C")
        assert len(weights) == len(
            node_ids
        ), "`weights` must have the same length as `node_ids`"

    # Get the segments (this will be a list of arrays of node indices)
    segments = _fastcore.generate_segments(parent_ix, weights=weights)

    # Map node indices back to IDs
    seg_ids = [node_ids[s] for s in segments]

    return seg_ids


def break_segments(node_ids, parent_ids):
    """Break neuron into linear segments connecting ends, branches and root.

    Parameters
    ----------
    node_ids :   (N, ) array
                 Array node IDs.
    parent_ids : (N, ) array
                 Array of parent IDs for each node. Root nodes' parents
                 must be -1.

    Returns
    -------
    segments :   list of arrays
                 Segments as list of arrays.

    Examples
    --------
    >>> import navis_fastcore as fastcore
    >>> import numpy as np
    >>> node_ids = np.arange(7)
    >>> parent_ids = np.array([-1, 0, 1, 2, 1, 4, 5])
    >>> fastcore.break_segments(node_ids, parent_ids)
    [array([1, 0]), array([3, 2]), array([6, 5, 4])]

    """
    # Convert parent IDs into indices
    parent_ix = _ids_to_indices(node_ids, parent_ids)

    # Get the segments (this will be a list of arrays of node indices)
    segments = _fastcore.break_segments(parent_ix)

    # Map node indices back to IDs
    seg_ids = [node_ids[s] for s in segments]

    return seg_ids


def segment_coords(
    node_ids,
    parent_ids,
    coords,
    weights=None,
    node_colors=None,
):
    """Generate coordinates for linear segments.

    This is useful for plotting the skeleton of a neuron.

    Parameters
    ----------
    node_ids :      (N, ) array
                    Array node IDs.
    parent_ids :    (N, ) array
                    Array of parent IDs for each node. Root nodes' parents
                    must be -1.
    coords :        (N, 3) array
                    Array of coordinates for each node.
    node_colors :   (N, ) numpy.ndarray, optional
                    A color for each node in `node_ids`. If provided, will
                    also return a list of colors sorted to match coordinates.

    Returns
    -------
    seg_coords :    list of arrays
    colors :        list of colors
                    If `node_colors` provided will return a copy of it sorted
                    to match `seg_coords`.

    Examples
    --------
    >>> import navis_fastcore as fastcore
    >>> import numpy as np
    >>> node_ids = np.arange(7)
    >>> parent_ids = np.array([-1, 0, 1, 2, 1, 4, 5])
    >>> coords = np.random.rand(7, 3)
    >>> fastcore.segment_coords(node_ids, parent_ids, coords)
    [array([[5.30713899e-01, 8.26450947e-01, 2.46805326e-01],
            [1.54144332e-04, 9.07823578e-01, 3.20199043e-01],
            [6.64580597e-01, 3.23724555e-01, 3.18361918e-01],
            [7.16579499e-01, 8.65568868e-02, 7.15686948e-01],
            [5.94874740e-01, 5.95528161e-01, 8.14234930e-01]]),
    array([[0.47814894, 0.84468164, 0.2765942 ],
            [0.21748528, 0.36673489, 0.81449368],
            [0.7165795 , 0.08655689, 0.71568695]])]

    """
    # Convert parent IDs into indices
    parent_ix = _ids_to_indices(node_ids, parent_ids)

    if weights is not None:
        weights = np.asarray(weights, dtype=np.float32, order="C")
        assert len(weights) == len(
            node_ids
        ), "`weights` must have the same length as `node_ids`"

    # Get the segments (this will be a list of arrays of node indices)
    segments = _fastcore.generate_segments(parent_ix, weights=weights)

    # Translate into coordinates
    seg_coords = [coords[s] for s in segments]

    # Apply colors if provided
    if not isinstance(node_colors, type(None)):
        colors = [node_colors[s] for s in segments]

        return seg_coords, colors

    return seg_coords


def geodesic_matrix(
    node_ids,
    parent_ids,
    directed=False,
    sources=None,
    targets=None,
    weights=None,
):
    """Calculate geodesic ("along-the-arbor") distances.

    Notes
    -----
    Under-the-hood, this uses two different implementations depending on whether
    a full all-by-all or a partial (via `sources`/`targets`) matrix is requested.
    The partial implementation is faster and more memory efficient for small-ish
    subsets of nodes. However, for subsets that include a large portion of the
    nodes, it may be faster to calculate the full matrix and then subset it.

    Parameters
    ----------
    node_ids :   (N, ) array
                 Array of node IDs.
    parent_ids : (N, ) array
                 Array of parent IDs for each node. Root nodes' parents
                 must be -1.
    directed :   bool, optional
                 If ``True`` will only return distances in the direction of
                 the child -> parent (i.e. towards the root) relationship.
    sources :    iterable, optional
                 Source node IDs. If ``None`` all nodes are used as sources.
    targets :    iterable, optional
                 Target node IDs. If ``None`` all nodes are used as targets.
    weights :    (N, ) float32 array, optional
                 Array of distances for each child -> parent connection.
                 If ``None`` all node to node distances are set to 1.

    Returns
    -------
    matrix :    float32 (single) array
                Geodesic distances. Unreachable nodes are set to -1.

    Examples
    --------
    >>> import navis_fastcore as fastcore
    >>> import numpy as np
    >>> node_ids = np.arange(7)
    >>> parent_ids = np.array([-1, 0, 1, 2, 1, 4, 5])
    >>> fastcore.geodesic_matrix(node_ids, parent_ids)
    array([[0., 1., 2., 3., 2., 3., 4.],
           [1., 0., 1., 2., 1., 2., 3.],
           [2., 1., 0., 1., 2., 3., 4.],
           [3., 2., 1., 0., 3., 4., 5.],
           [2., 1., 2., 3., 0., 1., 2.],
           [3., 2., 3., 4., 1., 0., 1.],
           [4., 3., 4., 5., 2., 1., 0.]], dtype=float32)
    >>> fastcore.geodesic_matrix(
    ...     node_ids, parent_ids,
    ...     sources=[0, 1], targets=[5, 6]
    ...     )
    array([[3., 4.],
           [2., 3.]], dtype=float32)

    """
    # Convert parent IDs into indices
    parent_ix = _ids_to_indices(node_ids, parent_ids)

    if weights is not None:
        weights = np.asarray(weights, dtype=np.float32, order="C")
        assert len(weights) == len(
            node_ids
        ), "`weights` must have the same length as `node_ids`"

    # Translate sources and targets into indices (if provided)
    if sources is not None:
        sources = np.where(np.isin(node_ids, sources))[0].astype(np.int32)
        assert len(sources), "`sources` must not be empty"
    if targets is not None:
        targets = np.where(np.isin(node_ids, targets))[0].astype(np.int32)
        assert len(targets), "`targets` must not be empty"

    # Calculate distances
    dists = _fastcore.geodesic_distances(
        parent_ix,
        sources=sources,
        targets=targets,
        weights=weights,
        directed=directed,
    )

    return dists


def geodesic_pairs(
    node_ids,
    parent_ids,
    pairs,
    directed=False,
    weights=None,
):
    """Calculate geodesic ("along-the-arbor") distances between pairs of nodes.

    This uses a simple algorithm that calculates distances using brute force.
    It's fast because we parallelize the calculation of each pair of nodes.

    Parameters
    ----------
    node_ids :   (N, ) array
                 Array of node IDs.
    parent_ids : (N, ) array
                 Array of parent IDs for each node. Root nodes' parents
                 must be -1.
    pairs :      (N, 2) array
                 Pairs of node IDs for which to calculate distances.
    directed :   bool, optional
                 If ``True`` will only return distances in the direction of
                 the child -> parent (i.e. towards the root) relationship.
    weights :    (N, ) float32 array, optional
                 Array of distances for each child -> parent connection.
                 If ``None`` all node to node distances are set to 1.

    Returns
    -------
    matrix :    float32 (single) array
                Geodesic distances. Unreachable nodes are set to -1.

    Examples
    --------
    >>> import navis_fastcore as fastcore
    >>> import numpy as np
    >>> node_ids = np.arange(7)
    >>> parent_ids = np.array([-1, 0, 1, 2, 1, 4, 5])
    >>> pairs = np.array([(0, 1), (0, 2)])
    >>> fastcore.geodesic_pairs(node_ids, parent_ids, pairs)

    """
    # Convert parent IDs into indices
    parent_ix = _ids_to_indices(node_ids, parent_ids)

    pairs = np.asarray(pairs)
    assert pairs.ndim == 2 and pairs.shape[1] == 2, "`pairs` must be of shape (N, 2)"

    if weights is not None:
        weights = np.asarray(weights, dtype=np.float32, order="C")
        assert len(weights) == len(
            node_ids
        ), "`weights` must have the same length as `node_ids`"

    # Calculate distances
    dists = _fastcore.geodesic_pairs(
        parent_ix,
        pairs_source=_ids_to_indices(node_ids, pairs[:, 0]),
        pairs_target=_ids_to_indices(node_ids, pairs[:, 1]),
        weights=weights,
        directed=directed,
    )

    return dists


def connected_components(node_ids, parent_ids):
    """Get the connected components for this neuron.

    Parameters
    ----------
    node_ids :   (N, ) array
                 Array of node IDs.
    parent_ids : (N, ) array
                 Array of parent IDs for each node. Root nodes' parents
                 must be -1.

    Returns
    -------
    cc :        (N, ) int32 array
                For each node the node ID of its root (= connected component ID).

    Examples
    --------
    Fully connected neuron:

    >>> import navis_fastcore as fastcore
    >>> import numpy as np
    >>> node_ids = np.arange(7)
    >>> parent_ids = np.array([-1, 0, 1, 2, 1, 4, 5])
    >>> fastcore.connected_components(node_ids, parent_ids)
    array([0, 0, 0, 0, 0, 0, 0])

    Introduce a break:

    >>> parent_ids[4] = -1
    >>> fastcore.connected_components(node_ids, parent_ids)
    array([0, 0, 0, 0, 4, 4, 4])

    """
    # Convert parent IDs into indices
    parent_ix = _ids_to_indices(node_ids, parent_ids)

    # Get connected components - this returns indices, not node IDs
    cc = _fastcore.connected_components(parent_ix)

    # Return the root node ID for each node
    return node_ids[cc]


def synapse_flow_centrality(
    node_ids, parent_ids, presynapses, postsynapses, mode="sum"
):
    """Calculate synapse flow centrality for this neuron.

    Please note that this implementation currently produces slightly different
    results than the implementation in navis. I'm not sure why that is but the
    differences seem to be negligible.

    Parameters
    ----------
    node_ids :   (N, ) array
                 Array of int32 node IDs.
    parent_ids : (N, ) array
                 Array of parent IDs for each node. Root nodes' parents
                 must be -1.
    presynapses : (N, ) uint32 array
                 Array of number of presynapses associated with each node.
    postsynapses : (N, ) uint32 array
                 Array of number of postsynapses associated with each node.
    mode :       "centrifugal" | "centripetal" | "sum"
                 The mode to calculate the flow centrality. "centrifugal" will
                 calculate the flow from the root to the leaves, "centripetal"
                 will calculate the flow from the leaves to the root, and "sum"
                 will calculate the sum of both.


    Returns
    -------
    cc :        (N, ) uint32 array
                Synapse flow centrality for each node.

    """
    # Convert parent IDs into indices
    parent_ix = _ids_to_indices(node_ids, parent_ids)

    # Make sure we have the correct data types and order
    presynapses = presynapses.astype(np.uint32, order="C", copy=False)
    postsynapses = postsynapses.astype(np.uint32, order="C", copy=False)

    assert len(presynapses) == len(postsynapses) == len(node_ids)

    # Get connected components - this returns indices, not node IDs
    flow = _fastcore.synapse_flow_centrality(parent_ix, presynapses, postsynapses, mode)

    # Return the flow for each node
    return flow


def parent_dist(node_ids, parent_ids, xyz, root_dist=None) -> None:
    """Get child->parent distances for skeleton nodes.

    Parameters
    ----------
    node_ids :   (N, ) array
                 Array of int32 node IDs.
    parent_ids : (N, ) array
                 Array of parent IDs for each node. Root nodes' parents
                 must be -1.
    xyz :        (N, 3) array
                 Array of coordinates for each node.
    root_dist :  int | None
                 ``parent_dist`` for the root's row. Set to ``None``, to leave
                 at ``NaN`` or e.g. to ``0`` to set to 0.

    Returns
    -------
    np.ndarray
                 Array with distances in same order and size as node table.

    """
    # Note: this function is effectively a copy of the one in navis with the
    # main difference being that it uses the fastcore implementation of
    # _ids_to_indices which is ~5X faster than the pandas-based version
    # in navis. Consider using this function for cable length calculations
    # instead of the graph-based one.

    # Convert parent IDs into indices
    parent_ix = _ids_to_indices(node_ids, parent_ids)
    not_root = parent_ix >= 0

    # Calculate distances
    w = np.full(len(parent_ix), root_dist, dtype=np.float32)
    w[not_root] = np.sqrt(((xyz[not_root] - xyz[parent_ix[not_root]]) ** 2).sum(axis=1))

    return w


def _ids_to_indices(node_ids, to_map):
    """Map node IDs to node indices.

    Parameters
    ----------
    node_ids :  (N, )
                Array of node IDs.
    to_map :    (N, )
                Array of IDs to map to indices. Root nodes' parents
                must be -1.

    Returns
    -------
    parent_ix : (N, ) int32 array
                Array with parent indices for each node.

    """
    # Some initial sanity checks
    node_ids = np.asanyarray(node_ids)
    to_map = np.asanyarray(to_map)
    assert node_ids.ndim == 1 and to_map.ndim == 1

    # Make sure node and parent IDs have the same dtype and downcast if necessary
    if node_ids.dtype != to_map.dtype:
        if node_ids.dtype == np.int32:
            to_map = to_map.astype(np.int32)
        elif node_ids.dtype == np.int64:
            node_ids = node_ids.astype(np.int32)
        else:
            raise ValueError("IDs must be int32 or int64")

    # Dispatch the correct function
    if node_ids.dtype == np.int32:
        return _fastcore.node_indices_32(node_ids, to_map)
    elif node_ids.dtype == np.int64:
        return _fastcore.node_indices_64(node_ids, to_map)
    else:
        raise ValueError("IDs must be int32 or int64")


def prune_twigs(node_ids, parent_ids, threshold, weights=None):
    """Prune twigs shorter than a given threshold.

    Parameters
    ----------
    node_ids :   (N, ) array
                 Array node IDs.
    parent_ids : (N, ) array
                 Array of parent IDs for each node. Root nodes' parents
                 must be -1.
    threshold :  float
                 Twigs shorter than this threshold will be pruned.
    weights :    (N, ) float32 array, optional
                 Array of distances for each child -> parent connection.
                 If ``None`` all node-to-node distances are set to 1.

    Returns
    -------
    keep :       (M, ) integer array
                 Node IDs to keep.

    Examples
    --------
    >>> import navis_fastcore as fastcore
    >>> import numpy as np
    >>> node_ids = np.arange(7)
    >>> parent_ids = np.array([-1, 0, 1, 2, 1, 4, 5])
    >>> fastcore.prune_twigs(node_ids, parent_ids, 2)
    array([0, 1, 4, 5, 6])

    """
    # Convert parent IDs into indices
    parent_ix = _ids_to_indices(node_ids, parent_ids)

    # Make sure weights are float32
    if weights is not None:
        weights = np.asarray(weights, dtype=np.float32, order="C")
        assert len(weights) == len(
            node_ids
        ), "`weights` must have the same length as `node_ids`"

    # Get the segments (this will be a list of arrays of node indices)
    keep_idx = _fastcore.prune_twigs(parent_ix, threshold, weights=weights)

    # Map node indices back to IDs
    return node_ids[keep_idx]


def strahler_index(
    node_ids, parent_ids, method="standard", to_ignore=None, min_twig_size=None
):
    """Calculcate Strahler Index.

    Parameters
    ----------
    node_ids :          (N, ) array
                        Array node IDs.
    parent_ids :        (N, ) array
                        Array of parent IDs for each node. Root nodes' parents
                        must be -1.
    method :            'standard' | 'greedy', optional
                        Method used to calculate Strahler indices: 'standard'
                        will use the method described above; 'greedy' will
                        always increase the index at converging branches
                        whether these branches have the same index or not.
    to_ignore :         iterable, optional
                        List of node IDs to ignore. Must be the FIRST node
                        of the branch. Excluded branches will not contribute
                        to Strahler index calculations and instead be assigned
                        the SI of their parent branch.
    min_twig_size :     int, optional
                        If provided, will ignore twigs with fewer nodes than
                        this. Instead, they will be assigned the SI of their
                        parent branch.

    Returns
    -------
    strahler_index :    (N, ) int array
                        Strahler Index for each node.

    Examples
    --------
    >>> import navis_fastcore as fastcore
    >>> import numpy as np
    >>> node_ids = np.arange(8)
    >>> parent_ids = np.array([-1, 0, 1, 2, 1, 4, 5, 5])
    >>> fastcore.strahler_index(node_ids, parent_ids)
    array([2, 2, 1, 1, 2, 2, 1, 1], dtype=int32)

    """
    # Convert parent IDs into indices
    parent_ix = _ids_to_indices(node_ids, parent_ids)

    # Convert to_ignore to indices
    if to_ignore is not None:
        to_ignore = np.where(np.isin(node_ids, to_ignore))[0].astype(np.int32)

    # Convert min_twig_size to int32
    if min_twig_size is not None:
        min_twig_size = np.int32(min_twig_size)

    # Get the Strahler indices
    strahler_index = _fastcore.strahler_index(
        parent_ix, min_twig_size=min_twig_size, to_ignore=to_ignore, method=method
    )

    # Map node indices back to IDs
    return strahler_index


def classify_nodes(node_ids, parent_ids):
    """Classify nodes.

    Note to self: this function is not significantly faster than the
    pandas/numpy-based implementation in navis which is already pretty
    fast. May need to test on larger neurons to see if there is a difference.

    Parameters
    ----------
    node_ids :          (N, ) array
                        Array node IDs.
    parent_ids :        (N, ) array
                        Array of parent IDs for each node. Root nodes' parents
                        must be -1.

    Returns
    -------
    node_type :         (N, ) integer array
                        Node types:
                         - 0: root
                         - 1: leaf
                         - 2: branch point
                         - 3: slab (intermediate node)

    Examples
    --------
    >>> import navis_fastcore as fastcore
    >>> import numpy as np
    >>> node_ids = np.arange(8)
    >>> parent_ids = np.array([-1, 0, 1, 2, 1, 4, 5, 5])
    >>> fastcore.classify_nodes(node_ids, parent_ids)
    array([2, 2, 1, 1, 2, 2, 1, 1], dtype=int32)

    """
    # Convert parent IDs into indices
    parent_ix = _ids_to_indices(node_ids, parent_ids)

    return _fastcore.classify_nodes(parent_ix)



import pytest
import time

import navis_fastcore as fastcore
import numpy as np
import pandas as pd

from scipy.spatial import cKDTree as KDTree
from pathlib import Path
from collections import namedtuple

Dotprop = namedtuple("Dotprop", ["points", "vect"])

# Set random state
np.random.seed(0)


def _load_swc(file="722817260.swc", dtype=np.int64, synapses=False):
    fp = Path(__file__).parent / file
    swc = pd.read_csv(fp, comment="#", header=None, sep=" ")
    nodes = swc[0].values.astype(dtype)
    coords = swc[[2, 3, 4]].values.astype(
        np.float64
    )  # the data type for this does not matter
    parents = swc[6].values.astype(dtype)

    if synapses:
        connectors = pd.read_csv(Path(__file__).parent / file.replace(".swc", ".csv"))
        presynapses = (
            connectors[connectors.type == "pre"]
            .node_id.value_counts()
            .reindex(nodes)
            .fillna(0)
            .values.astype(np.uint32, order="C", copy=False)
        )
        postsynapses = (
            connectors[connectors.type == "post"]
            .node_id.value_counts()
            .reindex(nodes)
            .fillna(0)
            .values.astype(np.uint32, order="C", copy=False)
        )
        return nodes, parents, coords, presynapses, postsynapses

    return nodes, parents, coords


# Number of nodes in the default test neuron
N_NODES = len(_load_swc()[0])


def swc64():
    return _load_swc(dtype=np.int32, synapses=False)


def swc32():
    return _load_swc(dtype=np.int64, synapses=False)


@pytest.fixture
def coords():
    return _load_swc()[2]


@pytest.fixture
def swc_with_synapses():
    return _load_swc(synapses=True)


def _node_indices_py(A, B):
    """Reference Python implementation of _node_indices."""
    ix_dict = dict(zip(B, np.arange(len(B))))

    return np.array([ix_dict.get(p, -1) for p in A])


@pytest.mark.parametrize("swc", [swc32(), swc64()])
def test_node_indices(swc):
    nodes, parents, _ = swc
    start = time.time()
    indices = fastcore.dag._ids_to_indices(nodes, parents)
    dur = time.time() - start

    start = time.time()
    indices_py = _node_indices_py(parents, nodes)
    dur_py = time.time() - start

    assert all(indices == indices_py)

    print("Indices:", indices)
    print(f"Timing: {dur:.4f}s (compared to {dur_py:.4f}s in pure numpy Python)")


@pytest.mark.parametrize("swc", [swc32(), swc64()])
@pytest.mark.parametrize("weights", [None, np.random.rand(N_NODES)])
def test_generate_segments(swc, weights):
    nodes, parents, _ = swc
    start = time.time()
    segments = fastcore.generate_segments(nodes, parents, weights=weights)
    dur = time.time() - start

    # print("Segments:", segments)
    print(type(segments), type(segments[0]))
    print(f"Timing: {dur:.4f}s")


@pytest.mark.parametrize("swc", [swc32(), swc64()])
def test_break_segments(swc):
    nodes, parents, _ = swc
    start = time.time()
    segments = fastcore.break_segments(nodes, parents)
    dur = time.time() - start

    print("Broken segments:", segments)
    print(f"Timing: {dur:.4f}s")


@pytest.mark.parametrize("swc", [swc32(), swc64()])
@pytest.mark.parametrize("node_colors", [None, np.random.rand(N_NODES)])
def test_segment_coords(swc, node_colors):
    nodes, parents, coords = swc
    start = time.time()
    coords = fastcore.segment_coords(nodes, parents, coords, node_colors=node_colors)
    dur = time.time() - start

    print("Segment coordinates:", coords)
    print(f"Timing: {dur:.4f}s")


@pytest.mark.parametrize("swc", [swc32(), swc64()])
@pytest.mark.parametrize("directed", [True, False])
@pytest.mark.parametrize("sources", [None, [0, 1, 2]])
@pytest.mark.parametrize("targets", [None, [0, 1, 2]])
@pytest.mark.parametrize("weights", [None, np.random.rand(N_NODES)])
def test_geodesic_distance(swc, directed, sources, targets, weights):
    nodes, parents, _ = swc
    start = time.time()
    dists = fastcore.geodesic_matrix(
        nodes,
        parents,
        directed=directed,
        sources=sources,
        targets=targets,
        weights=weights,
    )
    dur = time.time() - start

    print("Distances:", dists)
    print(f"Timing: {dur:.4f}s")


@pytest.mark.parametrize("swc", [swc32(), swc64()])
@pytest.mark.parametrize("directed", [True, False])
@pytest.mark.parametrize("weights", [None, np.random.rand(N_NODES)])
def test_geodesic_pairs(swc, directed, weights):
    nodes, parents, _ = swc
    pairs = np.vstack(
        (
            np.random.choice(nodes, 200),
            np.random.choice(nodes, 200),
        )
    ).T
    start = time.time()
    dists = fastcore.geodesic_pairs(
        nodes,
        parents,
        pairs=pairs,
        directed=directed,
        weights=weights,
    )
    dur = time.time() - start

    print("Distances:", dists)
    print(f"Timing: {dur:.4f}s")


@pytest.mark.parametrize("mode", ["centrifugal", "centripetal", "sum"])
def test_synapse_flow_centrality(swc_with_synapses, mode):
    nodes, parents, _, presynapses, postsynapses = swc_with_synapses
    start = time.time()
    cent = fastcore.synapse_flow_centrality(
        nodes, parents, presynapses, postsynapses, mode=mode
    )
    dur = time.time() - start

    print("Synapse flows:", cent)
    print(f"Timing: {dur:.4f}s")


@pytest.mark.parametrize("swc", [swc32(), swc64()])
@pytest.mark.parametrize("root_dist", [None, 0])
def test_parent_dist(swc, coords, root_dist):
    nodes, parents, _ = swc
    start = time.time()
    dists = fastcore.dag.parent_dist(nodes, parents, coords, root_dist)
    dur = time.time() - start

    print("Parent distances:", dists)
    print(f"Timing: {dur:.4f}s")


@pytest.mark.parametrize("swc", [swc32(), swc64()])
def test_connected_components(swc):
    nodes, parents, _ = swc
    start = time.time()
    cc = fastcore.connected_components(nodes, parents)
    dur = time.time() - start

    print("# of connected components:", len(np.unique(cc)))
    # print(type(segments), type(segments[0]))
    print(f"Timing: {dur:.4f}s")


@pytest.mark.parametrize("swc", [swc32(), swc64()])
@pytest.mark.parametrize("threshold", [5, 10])
@pytest.mark.parametrize("weights", [None, np.random.rand(N_NODES)])
def test_prune_twigs(swc, threshold, weights):
    nodes, parents, _ = swc
    start = time.time()
    pruned = fastcore.prune_twigs(nodes, parents, threshold=threshold, weights=weights)
    dur = time.time() - start

    print("Pruned nodes:", pruned)
    print(f"Timing: {dur:.4f}s")


@pytest.mark.parametrize("swc", [swc32(), swc64()])
@pytest.mark.parametrize("method", ["standard", "greedy"])
@pytest.mark.parametrize("min_twig_size", [None, 5])
def test_strahler_index(swc, method, min_twig_size):
    nodes, parents, _ = swc
    start = time.time()
    si = fastcore.strahler_index(
        nodes, parents, method=method, min_twig_size=min_twig_size
    )
    dur = time.time() - start

    print("Strahler indices:", si)
    print(f"Timing: {dur:.4f}s")


@pytest.mark.parametrize("swc", [swc32(), swc64()])
def test_classify_nodes(swc):
    nodes, parents, _ = swc
    start = time.time()
    types = fastcore.classify_nodes(nodes, parents)
    dur = time.time() - start

    print("Node types:", types)
    print(f"Timing: {dur:.4f}s")


def test_nblast_single():
    fp = Path(__file__).parent / "722817260.swc"
    swc1 = pd.read_csv(fp, comment="#", header=None, sep=" ")
    xyz1 = swc1[[2, 3, 4]].values.astype(np.float64) / 1000
    vec1 = calculate_tangent_vectors(xyz1, k=5)

    fp = Path(__file__).parent / "754534424.swc"
    swc2 = pd.read_csv(fp, comment="#", header=None, sep=" ")
    xyz2 = swc2[[2, 3, 4]].values.astype(np.float64) / 1000
    vec2 = calculate_tangent_vectors(xyz2, k=5)

    start = time.time()
    score = fastcore._fastcore.nblast_single(xyz1, vec1, xyz2, vec2, parallel=True)
    dur = time.time() - start

    print(f"NBLAST score: {score} ({dur:.4f}s)")


def calculate_tangent_vectors(points, k):
    """Calculate tangent vectors.

    Parameters
    ----------
    k :         int
                Number of nearest neighbours to use for tangent vector
                calculation.

    Returns
    -------
    Dotprops
                Only if ``inplace=False``.

    """
    # Create the KDTree and get the k-nearest neighbors for each point
    dist, ix = KDTree(points).query(points, k=k)

    # Get points: array of (N, k, 3)
    pt = points[ix]

    # Generate centers for each cloud of k nearest neighbors
    centers = np.mean(pt, axis=1)

    # Generate vector from center
    cpt = pt - centers.reshape((pt.shape[0], 1, 3))

    # Get inertia (N, 3, 3)
    inertia = cpt.transpose((0, 2, 1)) @ cpt

    # Extract vector and alpha
    u, s, vh = np.linalg.svd(inertia)
    vect = vh[:, 0, :]
    # alpha = (s[:, 0] - s[:, 1]) / np.sum(s, axis=1)

    return vect


if __name__ == "__main__":
    print("Done")
